#include <stdlib.h>
/* #include <malloc.h>*/
#include <memory.h>
#include <string.h>
#include <stdio.h>

#include "ams.h"
#include "acomm.h"
#include "netapi.h"
#include "netlib.h"
#include "sendfun.h"
#include "defs.h"
#include "amisc.h"


/* 
 * An ALICE communicator is a computational context of one or several
 * processors or nodes. Each node is identified by a hostname and a port
 * number 
 */

/* 
 * Global variables that I will get rid off them one of these days
 */

/* Communicators' Table */
COMM_STRUCT *COMM_TBL[MAX_PUB][MAX_COMM];

#ifdef AMS_SERVER
/* Publisher */
AMS_Pub PUBLISHER;
#endif

/* Published Communicators */
int PUBCOMM_TBL[MAX_COMM][MAX_COMM];

/* Global variables used in the client (accessor) */
char * COMM_LIST[MAX_PUB][MAX_COMM];

/* Current Publisher's id */
int pub_id = 0;

/* Publisher's id initialized to NULL */
char * PUBLISHER_ID[MAX_PUB] = {NULL};

#ifndef AMS_CLIENT
/* To synchronize the communicator table */
Mutex comm_pub_mutex;
#endif

/* AMS Error message defined in accpub.c */
extern char ams_msg[];

/*@
        AMSP_New_Host_Node - Creates a communicator and initialize it with some default values
        
        Output Parameters:
                err_code - Error code return.

@*/
HOST_NODE *AMSP_New_Host_Node(int *err)
{
    HOST_NODE *host_node = (HOST_NODE *) malloc(sizeof(HOST_NODE));
    if (host_node == NULL) {
        *err = AMS_ERROR_INSUFFICIENT_MEMORY;
        return NULL;
    }

    host_node->port = -1;
    host_node->prev = NULL;
    host_node->next = NULL;
    memset(host_node->hostname, 0, 255);

    *err = AMS_ERROR_NONE;
    return host_node;
}

/*@
        Del_Host_Node - Frees the resources associated with the node
        
        Input Parameters:
                host_node - A node within the communicator

@*/
int AMSP_Del_Host_Node(HOST_NODE *host)
{
    int err = Validate(AMSP_Valid_Host_Node,host);
    CheckErr(err);

    free(host);

    return AMS_ERROR_NONE;
}



/*@
        AMSP_Set_Host_Node - Sets the hostname and the port number of host_node.
        
        Input Parameters:
+			host_node - Node to be initialized.
.			host - hostname
+			port - port number for network connection


@*/
int AMSP_Set_Host_Node(HOST_NODE *host_node, char *host, int port)
{
    int err = Validate(AMSP_Valid_Host_Node, host_node);

    CheckErr(err);

    if (port <= 0 && port != -1)
        return AMS_ERROR_BAD_PORT_NUMBER;

    /* I doubt there will be hostname of 255 lenght */
    strncpy(host_node->hostname, host, 255);
    host_node->hostname[254] = 0;

    host_node->port = port;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Create_Host_Node - Creates host_node and sets its the hostname and port number.
        
        Input Parameters:
+			host_node - Node to be created and initialized.
.			host - hostname
+			port - port number for network connection


@*/
HOST_NODE * AMSP_Create_Host_Node(char *host, int port, int *err)
{
    HOST_NODE *host_node = AMSP_New_Host_Node(err);
    if (*err != AMS_ERROR_NONE)
        return NULL;

    if (port <= 0 && port != -1) {
        *err = AMS_ERROR_BAD_PORT_NUMBER;
        return NULL;
    }

    /* I doubt there will be hostname of 255 length */
    strncpy(host_node->hostname, host, 255);
    host_node->hostname[255] = 0;

    host_node->port = port;

    *err = AMS_ERROR_NONE;
    return host_node;
}


/*@
        AMSP_Find_Host_Comm_Struct - Given a hostname, the function checks if the
        host is in the Commnucation structure.

        Input Parameters:
+			com_str - A Communication structure
+			host - hostname to be searched
        Output Parameters:
+			found - Flag indicating the search status

@*/

int AMSP_Find_Host_Comm_Struct(COMM_STRUCT *com_str, char *host, HOST_NODE **found)
{       
    HOST_NODE *host_node;
    /* use "first names" until we have fully qualified hostnames */
    char our_firstname[MAXHOSTLEN+1], their_firstname[MAXHOSTLEN+1];
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
                
    *found = NULL;
    host_node = com_str->host_node;

    strcpy(our_firstname, host);
    strtok(our_firstname, ".");

    while (host_node->next) {
        strcpy(their_firstname, host_node->hostname);
        strtok(their_firstname, ".");
        if (strlen(their_firstname) == strlen(our_firstname)
            && !strcmp(their_firstname,  our_firstname)) {
            *found = host_node;
            break;
        }
        host_node = host_node->next;
    } /* Should terminate unless there is a bug */

    return AMS_ERROR_NONE;
}

#ifdef AMS_SERVER
int AMSP_Lock_Waiting_Room(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    AMSP_LockMutexWrite(&(com_str->waiting_room_lock));
    return AMS_ERROR_NONE;
}

int AMSP_UnLock_Waiting_Room(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    AMSP_RelMutexWrite(&(com_str->waiting_room_lock));
    return AMS_ERROR_NONE;
}


/*@ AMSP_Create_Appointment - Creates an appointment for a client thread

        Input Parameters:
                com_str - AMS communicator structure
                req - Request ID
                mem - AMS memory to which client desires access

        Output Parameters:
                appt - pointer to request's appointment
@*/

int AMSP_Create_Appointment(COMM_STRUCT *com_str, RequestID req, AMS_Memory mem, Appointment **appt)
{
	Appointment *tail;
	int reqsize, k;
	int err = Validate(AMSP_Valid_Comm_Struct, com_str);
	CheckErr(err);

	tail = com_str->waiting_room;

    *appt = (Appointment *) malloc(sizeof(Appointment));
	if (*appt == NULL)
		return AMS_ERROR_INSUFFICIENT_MEMORY;

    if(com_str->waiting_room == NULL)
        com_str->waiting_room = *appt;
    else {
        while(tail->next)
            tail = tail->next; /* Advance to the next appt */
        tail->next = *appt;
    }

    (*appt)->next = NULL; /* This is the last item on the list */

    err = AMSP_Size_RequestID(req, &reqsize);
    CheckErr(err);
    (*appt)->guest = (RequestID) malloc(reqsize);
    if((*appt)->guest == NULL)
        return AMS_ERROR_INSUFFICIENT_MEMORY;

    err = AMSP_Copy_RequestID((*appt)->guest, req);
    CheckErr(err);

    (*appt)->mem_wanted = mem;
	AMSP_CondInit(&((*appt)->cond));

    /* Denote indefinite appointment time */
    for(k = 0; k < STEP_BYTES; ++k)
        (*appt)->step[k] = (unsigned char) 255;

	return AMS_ERROR_NONE;

}

/*@ AMSP_Register_Guest - enter a new client request in the waiting room, finding its
                     appointment (creating it if necessary)

	Input Parameters:
+   ams - ALICE communicator
.   req - Request ID
+   mem - ALICE memory to which client desires access

    Output Parameters:
	appt - pointer to request's appointment

@*/

int AMSP_Register_Guest(AMS_Comm ams, RequestID req, AMS_Memory mem, Appointment **appt, int not)
{
    COMM_STRUCT *com_str;
    Appointment *tail;
    int found = 0, err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    tail = com_str->waiting_room;

    /* First try to find appointment, in case it pre-registered */
    if(tail != NULL) {
        if(AMSP_Equals_RequestID(tail->guest, req))
            if (tail->mem_wanted == mem)
				found = 1;

        while(tail->next && !found) {
            tail = tail->next;
            if(AMSP_Equals_RequestID(tail->guest, req))
                if (tail->mem_wanted == mem)
					found = 1;
        }
    }

    if(found) {
        *appt = tail;
        return AMS_ERROR_NONE;
    }

    /* Hasn't registered yet, create a new appointment */
	err = AMSP_Create_Appointment(com_str, req, mem, appt);
	CheckErr(err);

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Assign_Appt - record appointment time with waiting room, and tell the memory if necessary

        Input Parameters:
+       appt - appointment to schedule
+       step - time of appointment

        Be sure to lock the memory being requested before assigning an appointment time!
@*/

int AMSP_Assign_Appt(AMS_Comm ams, Appointment *appt, unsigned char *step)
{
    int err;
	
    if(appt == NULL)
        return AMS_ERROR_SYSTEM;

    memcpy((unsigned char *) appt->step, (const unsigned char *) step, STEP_BYTES);

    /* Memory is already locked when we do this */
    err = AMSP_Announce_Next_Guest(ams, appt->mem_wanted);
    CheckErr(err);

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Finish_Appt - delete appointment, tell memory the next time it's scheduled to stop

        Input Parameters:
+       ams - communicator
+       appt - appointment that just finished

@*/

int AMSP_Finish_Appt(AMS_Comm ams, Appointment *appt)
{
    COMM_STRUCT *com_str;
    Appointment *iter;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    /* Delete appointment */
        
    if(appt == com_str->waiting_room)
        com_str->waiting_room = appt->next;
    else {
        iter = com_str->waiting_room;
        while(iter->next) {
            if(iter->next == appt)
                iter->next = appt->next;
            if(iter->next)
                iter = iter->next;
        }
    }

    /* Who's next? */
    err = AMSP_Announce_Next_Guest(ams, appt->mem_wanted);
    CheckErr(err);

	AMSP_CondDestroy(&(appt->cond));
    free(appt->guest);
    free(appt);

    return AMS_ERROR_NONE;
}

int AMSP_Announce_Next_Guest(AMS_Comm ams, AMS_Memory mem)
{
    COMM_STRUCT *com_str;
    Appointment *iter;
    unsigned char next_time[STEP_BYTES];
    int k, err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    for(k = 0; k < STEP_BYTES; ++k)
        next_time[k] = (unsigned char) 255;

    for(iter = com_str->waiting_room; iter != NULL; iter = iter->next)
        if(iter->mem_wanted == mem && AMSP_StepGreaterThan(next_time, iter->step))
            memcpy(next_time, iter->step, STEP_BYTES);

    err = AMSP_Reserve_future_access(mem, (unsigned char *) next_time);
    CheckErr(err);

    return AMS_ERROR_NONE;
}

int AMSP_Wait_For_Appt_Step(AMS_Comm ams, Appointment *appt)
{
	COMM_STRUCT *com_str;
	int err = AMSP_Get_Comm_Tbl(ams, &com_str), t_flag;
    CheckErr(err);

	
    AMSP_Lock_Waiting_Room(ams);

    while(AMSP_StepUndef(appt->step)) {
		AMSP_CondWaitMutexWrite(&(appt->cond) ,&(com_str->waiting_room_lock), 0, &t_flag);
    }

    AMSP_UnLock_Waiting_Room(ams);

    return AMS_ERROR_NONE;
}
#endif

/*@
        AMSP_New_Comm_Struct - Creates an AMS communicator structure.
        
        Input Parameters:
			com_str - Structure to be initialized.

@*/
int AMSP_New_Comm_Struct(char *name, AMS_Comm_type ctype, AMS_Comm *ams)
{
    int i;
    COMM_STRUCT *com_str;
    HOST_NODE *tail_host;
    MEM_NODE *tail_mem;

    /* Init communication table if it is the first time */
    int err = AMSP_Init_Comm_Tbl();
    CheckErr(err);

    /* Lock the COMM_TBL table for modif */
    err = AMSP_Lock_TBL_Write_Comm_Struct();
    CheckErr(err);

    /* First, find an entry table */
    for (i=0; i<MAX_COMM; i++)
        if (COMM_TBL[pub_id][i] == NULL)
            break;
        
    if (i == MAX_COMM)
        return AMS_ERROR_COMM_TBL_FULL;

    /* Allocate the communication structure */
    com_str = (COMM_STRUCT *) malloc(sizeof(COMM_STRUCT));
    if (com_str == NULL)
        return AMS_ERROR_INSUFFICIENT_MEMORY;

    /* Give it a name */
    com_str->name = (char *) malloc((strlen(name)+1)*sizeof(char));
    if (com_str->name == NULL)
        return AMS_ERROR_INSUFFICIENT_MEMORY;
    strcpy(com_str->name, name);

    com_str->ctype = ctype;

    /* Create Tails of the list */
    tail_host = AMSP_New_Host_Node(&err);
    CheckErr(err);

    tail_mem = AMSP_New_Mem_Node(&err);
    CheckErr(err);


    /* Initialize the structure */

    com_str->nb_proc = 0;
    com_str->nb_mem = 0;
    com_str->TailHostNode = tail_host;
    com_str->TailMemNode = tail_mem;
    com_str->mem_node = tail_mem;
    com_str->host_node = tail_host;

#ifdef AMS_SERVER
    com_str->waiting_room = NULL;
    AMSP_NewMutex(&(com_str->waiting_room_lock));

    /* Create a mutex */
    AMSP_NewMutex(&(com_str->comm_mutex));

    /* Initialize the table */
    COMM_TBL[pub_id][i] = com_str;
    *ams = (AMS_Comm) i;
#else
    /* The entry is already know */
    CheckComm(*ams);
    COMM_TBL[pub_id][*ams] = com_str;
#endif

    /* Release the Lock on the Table */
    err = AMSP_UnLock_TBL_Write_Comm_Struct();
    CheckErr(err);

    return AMS_ERROR_NONE;
}


/*@
        AMSP_Del_Comm_Struct - a Deletes a communication structure and frees associated resources.
        
        Input Parameters:
			com_str - Structure to be initialized.

@*/
int AMSP_Del_Comm_Struct(COMM_STRUCT *com_str)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    err = AMSP_Del_Host_Node(com_str->TailHostNode);
    CheckErr(err);

    err = AMSP_Del_Mem_Node(com_str->TailMemNode);
    CheckErr(err);

#ifndef AMS_CLIENT
    AMSP_DelMutex(&(com_str->waiting_room_lock));
    AMSP_DelMutex(&(com_str->comm_mutex));
#endif

    free(com_str->name);
    free(com_str);

    return AMS_ERROR_NONE;
}



#ifdef AMS_SERVER
/*@
        AMSP_New_Pub_Arguments - Creates Publisher Start function arguments

        Input Parameters:
			port - Communication port number

        Output Parameter:
			parg - Address of a pointer to the arguments

@*/
int AMSP_New_Pub_Arguments(int port, Pub_arg **parg)
{
    if (port <= 0 && port != -1)
        return AMS_ERROR_BAD_ARGUMENT;

    if (parg == NULL)
        return AMS_ERROR_BAD_ARGUMENT;

    *parg = (Pub_arg *) malloc(sizeof(Pub_arg));
    if (*parg == NULL)
        return AMS_ERROR_INSUFFICIENT_MEMORY;

    /* Port number */
    (*parg)->port = port;

    (*parg)->care_about_port = 0;

    /* Status flag */
    (*parg)->alive = 0;

    /* Nb Thread connected */
    (*parg)->nb_client = 0;

    /* Status Mutex */
    AMSP_NewMutex(&( (*parg)->pub_mutex ));

    /* Nb of Connected Thread Mutex */
    AMSP_NewMutex(&((*parg)->nbthr_mutex));

	/* Condition variable to be signaled when the last thread is done */
	AMSP_CondInit(&((*parg)->cond));

    /* Condition variable to be signaled when the port number is set */
    AMSP_CondInit(&((*parg)->cond_port));

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Del_Pub_Arguments - Deletes the publisher start-funct arguments 

        Input Parameters:
			parg - Pointer the arguments
@*/
int AMSP_Del_Pub_Arguments(Pub_arg *parg)
{
    
    AMSP_DelMutex(&(parg->pub_mutex));
    AMSP_DelMutex(&(parg->nbthr_mutex));
	AMSP_CondDestroy(&(parg->cond));
    AMSP_CondDestroy(&(parg->cond_port));

    free(parg);
    return AMS_ERROR_NONE;
}

/*@
        AMSP_Set_Pub_Arguments_Comm_Struct - Sets the server startup function args

        Input Parameters:
			com_str - A Communication structure
            arg - Server-thread arguments
@*/
int AMSP_Set_Pub_Arguments_Comm_Struct(COMM_STRUCT *com_str, Pub_arg *arg)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    com_str->pub_arg = arg;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Set_MPI_Comm_Comm_Struct - Sets MPI_Comm member

        Input Parameters:
+			com_str - A Communication structure
+			mpi_comm - MPI Communicator

@*/
int AMSP_Set_MPI_Comm_Comm_Struct(COMM_STRUCT *com_str, MPI_Comm mpi_comm)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    com_str->mpi_comm = mpi_comm;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Get_MPI_Comm_Comm_Struct - Gets MPI_Comm member

        Input Parameter:
+			com_str - A Communication structure
		Output Parameter:
+			mpi_comm - Address of MPI Communicator

@*/

int AMSP_Get_MPI_Comm_Comm_Struct(COMM_STRUCT *com_str, MPI_Comm * mpi_comm)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    *mpi_comm = com_str->mpi_comm;

    return AMS_ERROR_NONE;
}

#endif

/*@
        AMSP_Set_Thread_ID_Comm_Struct - Sets the server-thread id

        Input Parameters:
+			com_str - A Communication structure
+			thr - Server-thread  id

@*/
#ifdef AMS_SERVER
int AMSP_Set_Thread_ID_Comm_Struct(COMM_STRUCT *com_str, Thread thr)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    com_str->thrid = thr;

    return AMS_ERROR_NONE;
}
#endif
/*@
        AMSP_Get_Thread_ID_Comm_Struct - Gets the server-thread id

        Input Parameters:
+			com_str - A Communication structure
+			thr - Server-thread  id

@*/
#ifdef AMS_SERVER
int AMSP_Get_Thread_ID_Comm_Struct(COMM_STRUCT *com_str, Thread *thr)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    *thr = com_str->thrid;

    return AMS_ERROR_NONE;
}
#endif
/*@
        AMSP_EndThread_Comm_Struct - Terminate the thread associated with a communicator

        Input Parameters:
			com_str - A Communication structure

@*/
#ifdef AMS_SERVER

int AMSP_EndThread_Comm_Struct(COMM_STRUCT *com_str)
{
    char host[MAXHOSTLEN+1], hosterr[255];
    int sfd, err1, err = Validate(AMSP_Valid_Comm_Struct, com_str), t_flag;
    CheckErr(err);

    /* server-thread should stop accepting new connections */
    AMSP_LockMutexWrite(&(com_str->pub_arg->pub_mutex));
    com_str->pub_arg->alive = 0;
    AMSP_RelMutexWrite(&(com_str->pub_arg->pub_mutex));

    /*
     * Any new pending request shoud be notified that the
     * the server-thread is not accepting new connections 
     */

    if (err = gethostname(host, MAXHOSTLEN))
        AMSP_err_quit("gethostname: %s", AMSP_host_err_str(hosterr, &err));

    /* Connect to myself if the communicator is still alive */
	err = AMSP_mem_net_open(host, AMS_SERVICE, com_str->pub_arg->port, &sfd, ams_msg);
	if (sfd < 0) {
		/* an error has occured */
		if (err != AMS_ERROR_CONNECT_REFUSED)
			return AMS_ERROR_NET_OPEN;
	}

	if (err != AMS_ERROR_CONNECT_REFUSED) {
		err1 = AMSP_send_rq(OP_SHUTDOWN, sfd);
		CheckErr(err1);

		err = AMSP_fsm_loop(OP_ATTACH_COMM, sfd, &err);
		AMSP_mem_net_close(sfd);
		if (err != AMS_ERROR_COMMUNICATOR_NO_LONGER_ALIVE)
			CheckErr(err);
		else
			err = AMS_ERROR_NONE;
	}

    /* 
	 * Wait until the last client thread is done. 
	 */

    AMSP_LockMutexRead(&(com_str->pub_arg->nbthr_mutex));
    while (com_str->pub_arg->nb_client) {
		AMSP_CondWaitMutexRead(&(com_str->pub_arg->cond), &(com_str->pub_arg->nbthr_mutex), 0, &t_flag);
    }
    AMSP_RelMutexRead(&(com_str->pub_arg->nbthr_mutex));

    /* 
     * Decrement the nubmer of communicators 
     */
    PUBLISHER.nb_com--;


    if (PUBLISHER.nb_com == 0) {

#ifdef AMS_LDAP
        /* If we are the last communicator, then delete the LDAP's publisher object */
        err = AMSP_Ldap_del_object(com_str->pub_arg->ld, PUBLISHER.dn);
        CheckErr(err);
        free(PUBLISHER.dn);
#endif
        PUBLISHER.alive = 0;
    }


    return AMS_ERROR_NONE;
}

/*@
    AMSP_Clean_up - Perform a clean up and delete AMS objects
@*/

void AMSP_Clean_up(void)
{
    int i;
    static called = 1;

    if (called == 1) {
        AMS_Print(" Performing AMS Clean up \n");
        called = 0;
    } else {
        AMS_Print(" Clean up failed \n");
        return;
    }

    /* Delete all the communicators */
    for (i=0;i<MAX_COMM;i++) {
        if (COMM_TBL[pub_id][i] != NULL)
            AMSP_EndThread_Comm_Struct(COMM_TBL[pub_id][i]);
    }

#ifdef AMS_LDAP
    /* delete the publisher if necessary */
    if (PUBLISHER.alive == 1) {
        AMSP_Ldap_del_object(PUBLISHER.ld, PUBLISHER.dn);
    }
#endif

}

/*@
    AMSP_Start_up - Perform a start up initialization
@*/

void AMSP_Start_up(void)
{
    /* Register clean up function */
    if (atexit(AMSP_Clean_up))
        AMSP_err_quit("AMS Start up failed \n");
}

/*@
    AMSP_Am_I_First_Host_Comm_Struct - are we the first host in our host_node list?
    Input Parameters:
+   com_str - Communicator Structure
    
    Output Paramters:
+   answer - pointer to an interger. 1 for yes, 0 for no.

@*/
int AMSP_Am_I_First_Host_Comm_Struct(COMM_STRUCT *com_str, int *answer)
{
    char ourname[MAXHOSTLEN+1], nickname[MAXHOSTLEN+1], hosterr[255];
    int ourport, err;

    if (err = gethostname(ourname, MAXHOSTLEN))
        AMSP_err_quit("gethostname: %s", AMSP_host_err_str(hosterr, &err));
    strtok(ourname, ".");
 
    strcpy(nickname, com_str->host_node->hostname);
    strtok(nickname, ".");

    ourport = com_str->pub_arg->port;

    *answer = 0;
    if(!strcmp(nickname, ourname) && ourport == com_str->host_node->port)
        *answer = 1;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Who_Is_Next_Host_Comm_Struct - get next host in host node list, or first if we're last
@*/
int AMSP_Who_Is_Next_Host_Comm_Struct(COMM_STRUCT *com_str, HOST_NODE **answer)
{
    char ourname[MAXHOSTLEN+1], nickname[MAXHOSTLEN+1], hosterr[255];
    int ourport, err, found;
    HOST_NODE *host_node;

    if (err = gethostname(ourname, MAXHOSTLEN))
        AMSP_err_quit("gethostname: %s", AMSP_host_err_str(hosterr, &err));
    strtok(ourname, ".");
 
    ourport = com_str->pub_arg->port;

    host_node = com_str->host_node;
    found = 0;
    *answer = NULL;
    while(host_node != com_str->TailHostNode) {
        strcpy(nickname, host_node->hostname);
        strtok(nickname, ".");
        if(!strcmp(nickname, ourname) && ourport == host_node->port) {
            found = 1;
            *answer = host_node->next;
        }
        host_node = host_node->next;
    }
        
    if(found == 0)
        return AMS_ERROR_SYSTEM;

    /* After last host, return to first */
    if(*answer == com_str->TailHostNode)
        *answer = com_str->host_node;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Get_Nb_Proc_Comm_Struct - get the number of processors
@*/
extern int AMSP_Get_Nb_Proc_Comm_Struct(COMM_STRUCT *com_str, int *nb)
{
    *nb = com_str->nb_proc;
    return AMS_ERROR_NONE;
}

#endif

/*@
        AMSP_Lock_TBL_Write_Comm_Struct - Lock the Communicators' table 
@*/
int AMSP_Lock_TBL_Write_Comm_Struct()
{
    /* This function will abort if there is an error */
#ifndef AMS_CLIENT
    AMSP_LockMutexWrite(&comm_pub_mutex);
#endif

    return AMS_ERROR_NONE;
}

int AMSP_Lock_TBL_Read_Comm_Struct()
{
    /* This function will abort if there is an error */
#ifndef AMS_CLIENT
    AMSP_LockMutexRead(&comm_pub_mutex);
#endif

    return AMS_ERROR_NONE;
}

/*@
    AMSP_UnLock_TBL_Write_Comm_Struct - Unlock the Communicators' table 
@*/
int AMSP_UnLock_TBL_Write_Comm_Struct()
{
    /* This function will abort if there is an error */
#ifndef AMS_CLIENT
    AMSP_RelMutexWrite(&comm_pub_mutex);
#endif

    return AMS_ERROR_NONE;
}

/*@
	AMSP_UnLock_TBL_Read_Comm_Struct - Unlock the Communicators' table 
@*/
int AMSP_UnLock_TBL_Read_Comm_Struct()
{
    /* This function will abort if there is an error */
#ifndef AMS_CLIENT
    AMSP_RelMutexRead(&comm_pub_mutex);
#endif

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Find_Mem_Node_Comm_Struct - Given a memory node, the function checks if the
        node is in the Communication structure.

        Input Parameters:
+			com_str - A Communication structure
+			mem - Memory node to be searched
        Output Parameters:
                mem_node - pointer to the node if found, NULL otherwise

@*/

int AMSP_Find_Mem_node_Comm_Struct(COMM_STRUCT *com_str, MEM_NODE *mem, MEM_NODE **mem_node)
{       
    MEM_NODE *node;
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
                
    err = Validate(AMSP_Valid_Mem_Node, mem);
    CheckErr(err);

    *mem_node = NULL;

    node = com_str->mem_node;

    while (node->next) {
        switch ((err = AMSP_Compare_Mem_Node(mem, node))) {
        case AMS_IDENTICAL_MEMORY:
            *mem_node = node;
            break;
        case AMS_DIFFERENT_MEMORY:
            break;
        default:
            return err;
        }

        if (*mem_node)
            break;

        node = node->next;
    } /* Should terminate unless I am a dummy */

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Find_Mem_name_Comm_Struct - Given a memory name, the function checks if the
        memory is in the Commnucation structure (has been published).

        Input Parameters:
+			com_str - A Communication structure
+			name - memory name to search
        Output Parameters:
+			mem - memory id
+			found - Flag indicating the search result

@*/

int AMSP_Find_Mem_name_Comm_Struct(COMM_STRUCT *com_str, char *name, AMS_Memory *mem, MEM_NODE **node)
{       
    MEM_NODE *mem_node;
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
                
    mem_node = com_str->mem_node;

    while (mem_node->next) {
        if (!strcmp(name, mem_node->amem->name)) {
            *node = mem_node;
            *mem = mem_node->memid;
            break;
        }
        mem_node = mem_node->next;
    } /* Should terminate unless there is a bug */

    if (mem_node->next == NULL)
        *node = NULL;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Find_Mem_id_Comm_Struct - Given a memory id, the function checks if the
        memory id is in the Commnucation structure.

        Input Parameters:
+			com_str - A Communication structure
+			id - memory structure id to search
        Output Parameters:
+			node - Memory node (NULL if the id is not found)

@*/

int AMSP_Find_Mem_id_Comm_Struct(COMM_STRUCT *com_str, AMS_Memory id, MEM_NODE **node)
{       
    MEM_NODE *mem_node;
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
                
    mem_node = com_str->mem_node;

    while (mem_node->next) {
        if (mem_node->memid == id) {
            *node = mem_node;
            break;
        }
        mem_node = mem_node->next;
    } /* Should terminate unless there is a bug */

    if (mem_node->next == NULL)
        *node = NULL;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Init_Comm_Tbl - Initializes the communication table. This 
        function is not thread-safe.
        
@*/
int AMSP_Init_Comm_Tbl()
{
    int i;
    static int firstime = 1;

#ifndef AMS_CLIENT
    int err;
#endif

    if (firstime) {
        /* 
         * 007: there is a potential conflict if two threads
         * enter here at the same time. We will look at it later.
         */

        firstime = 0;
#ifndef AMS_CLIENT
        AMSP_NewMutex(&comm_pub_mutex);
        err = AMSP_Lock_TBL_Write_Comm_Struct();
        CheckErr(err);
#endif

        for (i=0; i<MAX_COMM; i++) {
            int j;
            for (j=0; j<MAX_PUB;j++) {

                COMM_TBL[j][i] = NULL;
                PUBCOMM_TBL[j][i] = -1;
            }
        }

#ifndef AMS_CLIENT
        err = AMSP_UnLock_TBL_Write_Comm_Struct();
        CheckErr(err);
#endif
    }

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Set_Comm_Tbl - Assign a value to a table entry
@*/

int AMSP_Set_Comm_Tbl(int id, COMM_STRUCT *com_str)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    CheckComm(id);

#ifndef AMS_CLIENT
    err = AMSP_Lock_TBL_Write_Comm_Struct();
    CheckErr(err);
#endif

    COMM_TBL[pub_id][id] = com_str;

#ifndef AMS_CLIENT
    err = AMSP_UnLock_TBL_Write_Comm_Struct();
    CheckErr(err);
#endif

    return err;
}


/*@
        AMSP_Get_Comm_Tbl - Get a value within the table
@*/

int AMSP_Get_Comm_Tbl(int id, COMM_STRUCT **com_str)
{
    int err = AMS_ERROR_NONE;

    CheckComm(id);

#ifndef AMS_CLIENT
    err = AMSP_Lock_TBL_Read_Comm_Struct();
    CheckErr(err);
#endif

    *com_str = COMM_TBL[pub_id][id];

#ifndef AMS_CLIENT
    err = AMSP_UnLock_TBL_Read_Comm_Struct();
    CheckErr(err);
#endif

    return err;
}


/*@
        AMSP_Publish_Comm_Struct - Connects to the daemon to publish the communicator
        
        Input Parameters:
+			com_str - Communication structure

@*/
int AMSP_Publish_Comm_Struct(AMS_Comm ams)
{
    int err;
    COMM_STRUCT *com_str;

    err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifndef AMS_CLIENT
    err = AMSP_Lock_TBL_Write_Comm_Struct();
    CheckErr(err);
#endif

    if (PUBCOMM_TBL[pub_id][ams] != -1)
        return AMS_ERROR_SYSTEM;

    PUBCOMM_TBL[pub_id][ams] = 1;

#ifndef AMS_CLIENT
    err = AMSP_UnLock_TBL_Write_Comm_Struct();
    CheckErr(err);
#endif

    return AMS_ERROR_NONE;
}

/*@
        AMSP_UnPublish_Comm_Struct - Deletes a communication structure from the table
        
        Input Parameters:
+			com_str - Coomunication structure

@*/
int AMSP_UnPublish_Comm_Struct(AMS_Comm ams)
{
    int err;
    COMM_STRUCT *com_str;

    err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifndef AMS_CLIENT
    /* Lock the table */
    err = AMSP_Lock_TBL_Write_Comm_Struct();
    CheckErr(err);
#endif

    /* Modify the table */
    PUBCOMM_TBL[pub_id][ams] = -1;

#ifndef AMS_CLIENT
    /* UnLock the table */
    err = AMSP_UnLock_TBL_Write_Comm_Struct();
    CheckErr(err);
#endif

    return AMS_ERROR_NONE;

}

/*@
        AMSP_Add_Host_Node_Comm_Struct - Attaches a host_node to a communication structure
        
        Input Parameters:
+			com_str - Communication structure
+			host_node - HOST_NODE to add

@*/
int AMSP_Add_Host_Node_Comm_Struct(COMM_STRUCT *com_str, HOST_NODE *host_node)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);

    /* Check the validity of the arguments */
    CheckErr(err);
                
    err = Validate(AMSP_Valid_Host_Node, host_node);
    CheckErr(err);
                
        
    /* Appends a new field at the end */
    if (com_str->host_node == com_str->TailHostNode) {
        /* Empty list, add the new field at the beginning */
        host_node->next = com_str->TailHostNode;
        host_node->prev = NULL;
        com_str->host_node = host_node;
        com_str->TailHostNode->prev = host_node;
    } else {
        /* Append at the end */
        host_node->next = com_str->TailHostNode;
        host_node->prev = com_str->TailHostNode->prev;
        host_node->prev->next = host_node;
        com_str->TailHostNode->prev = host_node;
    }

    com_str->nb_proc += 1;

    return AMS_ERROR_NONE;
}


/*@
        AMSP_Del_Host_Node_Comm_Struct - Deletes a host_node from a communication structure.
        
        Input Parameters:
+			com_str - Communication structure
+			host_node - Node to delete

@*/
int AMSP_Del_Host_Node_Comm_Struct(COMM_STRUCT *com_str, HOST_NODE *host_node)
{
    return AMS_ERROR_FUNCT_NOTIMPLEMENTED;
}

/*@
        AMSP_Del_Host_Node_Comm_Struct - Deletes a host_node from a communication structure.
        
        Input Parameters:
+			com_str - Communication structure
+			host_node - Node to delete

@*/
int AMSP_DelAll_Host_Comm_Struct(COMM_STRUCT *com_str)
{
    HOST_NODE *host;
    int err = Validate(AMSP_Valid_Comm_Struct,com_str), nb_host = 0;
    CheckErr(err);

    host = com_str->host_node;
    while (host->next) {
        host = host->next;
        err = AMSP_Del_Host_Node(host->prev);
        CheckErr(err);
        nb_host++;
    }

    /* Reconnect the tail */
    com_str->host_node = host; 

    /* Just Checking */
    if (nb_host != com_str->nb_proc)
        return AMS_ERROR_SYSTEM;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Add_Mem_Node_Comm_Struct - Attaches a mem_node to a communication structure
        
        Input Parameters:
+			com_str - Communication structure
+			mem_node - Memory node to add

@*/
int AMSP_Add_Mem_Node_Comm_Struct(COMM_STRUCT *com_str, MEM_NODE *mem_node)
{
    int err = Validate(AMSP_Valid_Comm_Struct,com_str);
    MEM_NODE *node;

    /* Check the validity of the arguments */
    CheckErr(err);

    err = Validate(AMSP_Valid_Mem_Node, mem_node);
    CheckErr(err);
        
    /* Make sure the mem_node is not already in the communicator */
    err = AMSP_Find_Mem_node_Comm_Struct(com_str, mem_node, &node);
    CheckErr(err);

    if (node)
        return AMS_ERROR_NODE_ALREADY_IN_COMM;

    /* Appends a new node at the end */
    if (com_str->mem_node == com_str->TailMemNode) {
        /* Empty list, add the new node at the beginning */
        mem_node->next = com_str->TailMemNode;
        mem_node->prev = NULL;
        com_str->mem_node = mem_node;
        com_str->TailMemNode->prev = mem_node;
    } else {
        /* Append at the end */
        mem_node->next = com_str->TailMemNode;
        mem_node->prev = com_str->TailMemNode->prev;
        mem_node->prev->next = mem_node;
        com_str->TailMemNode->prev = mem_node;
    }

    com_str->nb_mem += 1;

    return AMS_ERROR_NONE;
}


/*@
        AMSP_Del_Mem_Node_Comm_Struct - Deletes a mem_node from a communication structure.
        
        Input Parameters:
+			com_str - Communication structure
+			mem_node - Node to delete

@*/
int AMSP_Del_Mem_Node_Comm_Struct(COMM_STRUCT *com_str, MEM_NODE *mem_node)
{
    MEM_NODE *node;
    int err = Validate(AMSP_Valid_Comm_Struct, com_str), found = 0;
    CheckErr(err);

    /* Find the memory node */
    err = AMSP_Find_Mem_node_Comm_Struct(com_str, mem_node, &node);
    CheckErr(err);

    if (!node)
        return AMS_ERROR_MEMORY_NODE_NOT_IN_COMMUNICATOR;

    /* Delete from the beginning */
    if (node->prev == NULL) {
        com_str->mem_node = node->next;
        node->next->prev = NULL;
    } else {
        node->prev->next = node->next;
        node->next->prev = node->prev;
    }
    com_str->nb_mem -= 1;

    err = AMSP_Del_Mem_Node(node);
    CheckErr(err);

    return AMS_ERROR_NONE;
}

/*@
        AMSP_DelAll_Mem_Comm_Struct - Deletes all nodes from a communication structure.
        
        Input Parameters:
+			com_str - Communication structure

@*/
int AMSP_DelAll_Mem_Comm_Struct(COMM_STRUCT *com_str)
{

    MEM_NODE *mem;
    int err = Validate(AMSP_Valid_Comm_Struct, com_str), nb_mem = 0;
    CheckErr(err);

    mem = com_str->mem_node;
    while (mem->next) {
        mem = mem->next;
        err = AMSP_Del_Mem_Node(mem->prev);
        CheckErr(err);
        nb_mem++;
    }

    /* Reconnect the tail */
    com_str->mem_node = mem; 

    if (nb_mem != com_str->nb_mem)
        return AMS_ERROR_SYSTEM;

    return AMS_ERROR_NONE;
}

/*@
    AMSP_Get_Comm_type_Comm_Struct - Gets a communication type
        
    Input Parameter:
+   com_str - Communication structure
    Output Parameter:
+   ctype - Pointer to the communicator type (MPI_TYPE, NODE_TYPE)

@*/
int AMSP_Get_Comm_type_Comm_Struct(COMM_STRUCT *com_str, AMS_Comm_type *ctype)
{
    int err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    *ctype = com_str->ctype;

    return AMS_ERROR_NONE;
}


/* Interface to AMS_Comm functions */
int AMSP_New_AMS_Comm(char *name, AMS_Comm_type ctype, AMS_Comm *ams)
{
    return AMSP_New_Comm_Struct(name, ctype, ams);
}

/*
 * Given an AMS_Comm, this function returns a Comm Structure
 */
int AMSP_Get_Comm_struct_AMS_Comm(AMS_Comm ams, COMM_STRUCT **com_str)
{

    /* Lock the table */
    int err;
    CheckComm(ams);

    err = AMSP_Lock_TBL_Read_Comm_Struct();
    CheckErr(err);

    *com_str = COMM_TBL[pub_id][ams];

    /* UnLock the table */
    err = AMSP_UnLock_TBL_Read_Comm_Struct();
    CheckErr(err);

    return AMS_ERROR_NONE;
}

#ifdef AMS_SERVER
int AMSP_Set_Pub_Arguments_AMS_Comm(AMS_Comm ams, Pub_arg *arg)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    return AMSP_Set_Pub_Arguments_Comm_Struct(com_str, arg);
}

int AMSP_Set_Thread_ID_AMS_Comm(AMS_Comm ams, Thread thr)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    return AMSP_Set_Thread_ID_Comm_Struct(com_str, thr);
}

int AMSP_Get_Thread_ID_AMS_Comm(AMS_Comm ams, Thread *thr)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    return AMSP_Get_Thread_ID_Comm_Struct(com_str, thr);
}

int AMSP_Set_MPI_Comm_AMS_Comm(AMS_Comm ams, MPI_Comm mpi_comm)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    return AMSP_Set_MPI_Comm_Comm_Struct(com_str, mpi_comm);
}

int AMSP_Get_MPI_Comm_AMS_Comm(AMS_Comm ams, MPI_Comm *mpi_comm)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    return AMSP_Get_MPI_Comm_Comm_Struct(com_str, mpi_comm);
}

int AMSP_Lock_Read_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    AMSP_LockMutexRead(&(com_str->comm_mutex));

    return AMS_ERROR_NONE;
}

int AMSP_UnLock_Read_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    AMSP_RelMutexRead(&(com_str->comm_mutex));

    return AMS_ERROR_NONE;
}

int AMSP_Lock_Write_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    AMSP_LockMutexWrite(&(com_str->comm_mutex));

    return AMS_ERROR_NONE;
}

int AMSP_UnLock_Write_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    AMSP_RelMutexWrite(&(com_str->comm_mutex));

    return AMS_ERROR_NONE;
}

#endif

int AMSP_Publish_AMS_Comm(AMS_Comm ams)
{
    CheckComm(ams);
    return AMSP_Publish_Comm_Struct(ams);
}

int AMSP_UnPublish_AMS_Comm(AMS_Comm ams)
{
    CheckComm(ams);
    return AMSP_UnPublish_Comm_Struct(ams);
}
/*
 * Check if the communicator is currently published 
 */
/* NOTE: This function need not be called by a communicator thread! */
int AMSP_IsPublished_AMS_Comm(AMS_Comm ams, int lock, int *flag)
{

#ifdef AMS_SERVER
    int err;
    if (lock != AMS_DONOT_LOCK) {
        err = AMSP_Lock_TBL_Read_Comm_Struct();
        CheckErr(err);
    }

    if (PUBCOMM_TBL[pub_id][ams] == 1)
        *flag = TRUE;
    else
        *flag = FALSE;

    if (lock == AMS_RELEASE_LOCK || *flag == FALSE)
        if (lock != AMS_DONOT_LOCK)
            AMSP_UnLock_TBL_Read_Comm_Struct();
        
#endif
    return AMS_ERROR_NONE;

}

#ifndef AMS_CLIENT
int AMSP_EndThread_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
    return AMSP_EndThread_Comm_Struct(com_str);
}
#endif

int AMSP_Add_Host_Node_AMS_Comm(AMS_Comm ams, HOST_NODE *host)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifdef AMS_SERVER
    err = AMSP_Lock_Write_AMS_Comm(ams);
    CheckErr(err);
#endif

    err = AMSP_Add_Host_Node_Comm_Struct(com_str, host);

#ifdef AMS_SERVER
    AMSP_UnLock_Write_AMS_Comm(ams);
#endif

    return err;
}

int AMSP_Del_Host_Node_AMS_Comm(AMS_Comm ams, HOST_NODE *host)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifdef AMS_SERVER
    err = AMSP_Lock_Write_AMS_Comm(ams);
    CheckErr(err);
#endif

    err = AMSP_Del_Host_Node_Comm_Struct(com_str, host);

#ifdef AMS_SERVER
    AMSP_UnLock_Write_AMS_Comm(ams);
#endif

    return err;
}

int AMSP_Add_Mem_Node_AMS_Comm(AMS_Comm ams, MEM_NODE *mem)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifdef AMS_SERVER
    err = AMSP_Lock_Write_AMS_Comm(ams);
    CheckErr(err);
#endif

    err = AMSP_Add_Mem_Node_Comm_Struct(com_str, mem);

#ifdef AMS_SERVER
    AMSP_UnLock_Write_AMS_Comm(ams);
#endif

    return err;

}

int AMSP_Find_Mem_name_AMS_Comm(AMS_Comm ams, char *name, AMS_Memory *mem, MEM_NODE **mem_node)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
    return AMSP_Find_Mem_name_Comm_Struct(com_str, name, mem, mem_node);
}

int AMSP_Find_Mem_node_AMS_Comm(AMS_Comm ams, MEM_NODE *mem_node, MEM_NODE **node)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
    return AMSP_Find_Mem_node_Comm_Struct(com_str, mem_node, node);
}

int AMSP_Find_Mem_id_AMS_Comm(AMS_Comm ams, AMS_Memory mem, MEM_NODE **node)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
    return AMSP_Find_Mem_id_Comm_Struct(com_str, mem, node);
}

int AMSP_Del_Mem_Node_AMS_Comm(AMS_Comm ams, MEM_NODE *mem_node)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
    return AMSP_Del_Mem_Node_Comm_Struct(com_str, mem_node);
}

int AMSP_DelAll_Host_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);
    return      AMSP_DelAll_Host_Comm_Struct(com_str);
}

int AMSP_DelAll_Mem_AMS_Comm(AMS_Comm ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifdef AMS_SERVER
    err = AMSP_Lock_Write_AMS_Comm(ams);
    CheckErr(err);
#endif

    err = AMSP_DelAll_Mem_Comm_Struct(com_str);

#ifdef AMS_SERVER
    AMSP_UnLock_Write_AMS_Comm(ams);
#endif

    return err;
}
        
int AMSP_Del_AMS_Comm(ams)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    err = AMSP_Del_Comm_Struct(com_str);
    CheckErr(err);

    /* Lock the table */
    err = AMSP_Lock_TBL_Write_Comm_Struct();
    CheckErr(err);

    /* Modify the table */
    COMM_TBL[pub_id][ams] = NULL;

    /* UnLock the table */
    err = AMSP_UnLock_TBL_Write_Comm_Struct();
    CheckErr(err);

    return AMS_ERROR_NONE;
}

/*
 * Given a Communicator name, this function returns its id
 */
int AMSP_Get_AMS_Comm(char *name, AMS_Comm *ams)
{
    int id, published = 0, err;
    COMM_STRUCT *com_str;

    /* Lock the table */
    err = AMSP_Lock_TBL_Read_Comm_Struct();
    CheckErr(err);

    for (id = 0; id < MAX_COMM; id++) {
        if (com_str = COMM_TBL[pub_id][id])
            if (!strcmp(com_str->name, name)) {
                *ams = id;
                published = PUBCOMM_TBL[pub_id][id];
                break;
            }
    }

    /* UnLock the table */
    err = AMSP_UnLock_TBL_Read_Comm_Struct();
    CheckErr(err);

    if (id == MAX_COMM)
        return AMS_ERROR_COMM_NAME_NOT_FOUND;
        
    if (published == 0)
        return AMS_ERROR_COMM_NOT_PUBLISHED;

    return AMS_ERROR_NONE;
}

/*
 * Given a communicator id, this function returns its type (MPI or NODE)
 *
 */
int AMSP_Get_Comm_type_AMS_Comm(AMS_Comm ams, AMS_Comm_type *ctype)
{
    COMM_STRUCT *com_str;

    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    return AMSP_Get_Comm_type_Comm_Struct(com_str, ctype);
}

/*
 * Given a communicator id, this function returns the list
 * of memory names attached to the communicator
 */
int AMSP_Get_Memory_list_AMS_Comm(AMS_Comm ams, char **mem_list, int *list_len)
{

    COMM_STRUCT *com_str;
    MEM_NODE *mem_node;

    int flag, len, err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

#ifdef AMS_SERVER
    err = AMSP_Lock_Read_AMS_Comm(ams);
    CheckErr(err);
#endif

    /* Length */
    mem_node = com_str->mem_node;
    len = 0;
    while(mem_node->next) {
        if (mem_node->amem == NULL)
            return AMS_ERROR_SYSTEM;

        /* Do not lock. Already locked */
        err = AMSP_IsPublished_AMS_Memory(&(mem_node->memid), 1, AMS_DONOT_LOCK, &flag);
        CheckErr(err);

        if (flag)
            len += strlen(mem_node->amem->name) + 1; /* +1 for zero terminator */
        mem_node = mem_node->next;
    }

    /* Memory allocation */
    *mem_list = (char *) malloc(len * sizeof(char));
    if (*mem_list == NULL)
        return AMS_ERROR_INSUFFICIENT_MEMORY;

    /* Data */
    mem_node = com_str->mem_node;
    *list_len = 0;
    while(mem_node->next) {
        if (mem_node->amem == NULL)
            return AMS_ERROR_SYSTEM;

        /* Do not lock. Already locked */
        err = AMSP_IsPublished_AMS_Memory(&(mem_node->memid), 1, AMS_DONOT_LOCK, &flag);
        CheckErr(err);

        if (flag) {
            strcpy(*mem_list + *list_len, mem_node->amem->name);
            *list_len += strlen(mem_node->amem->name) + 1; /* +1 for zero terminator */
        }
        mem_node = mem_node->next;
    }


#ifdef AMS_SERVER
    AMSP_UnLock_Read_AMS_Comm(ams);
#endif

    if (len != *list_len)
        AMSP_err_quit("Get_Memory_list_AMS_Comm: Length calculated differs from length sent");

    return AMS_ERROR_NONE;
}

/*
 * Given a communicator id, this function returns the list
 * of host names and port# used within the communicator
 */
int AMSP_Get_Host_list_AMS_Comm(AMS_Comm ams, char **host_list, int *list_len)
{

    COMM_STRUCT *com_str;
    HOST_NODE *host_node;
    char tmp[20];

    int flag, err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    /* Necessary since we're the publisher thread */
    err = AMSP_IsPublished_AMS_Comm(ams, AMS_RELEASE_LOCK, &flag);
    CheckErr(err);

    if (!flag)
        return AMS_ERROR_COMMUNICATOR_NO_LONGER_ALIVE;

#ifdef AMS_SERVER
    err = AMSP_Lock_Read_AMS_Comm(ams);
    CheckErr(err);
#endif

    /* Length */
    host_node = com_str->host_node;
    *list_len = 0;
    while(host_node->next) {
        sprintf(tmp, "%d", host_node->port);
        *list_len += strlen(host_node->hostname) + strlen(tmp) + 2; /* +2 for zero terminator and "|" */
        host_node = host_node->next;
    }

    /* Allocated Memory: the caller is responsible of freeing it */
    *host_list = (char *)malloc(*list_len * sizeof(char));
    if (*host_list == NULL)
        return AMS_ERROR_INSUFFICIENT_MEMORY;

    /* Data */
    host_node = com_str->host_node;
    *list_len = 0;
    while(host_node->next) {
        strcpy(*host_list + *list_len, host_node->hostname);
        strcat(*host_list + *list_len, "|");
        sprintf(tmp, "%d", host_node->port);
        strcat(*host_list + *list_len, tmp);
        *list_len += strlen(host_node->hostname) + strlen(tmp) + 2; /* +2 for zero terminator and "|" */
        host_node = host_node->next;
    }

#ifdef AMS_SERVER
    AMSP_UnLock_Read_AMS_Comm(ams);
#endif

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Set_publisher_id - Find the hostname and port number in the Publishers'
        table and sets the current publisher id to the the index that was found.

        Input parameters:
+			host - Publisher's host name
+			port - Publisher's port number

@*/

int AMSP_Set_publisher_id(char *host, int port)
{
    int i;
    char s_port[20];

    /* Convert port number to string */
    sprintf(s_port, "%d", port);

    for (i=0; i<MAX_PUB; i++) {
        if (PUBLISHER_ID[i])
            if (strstr(PUBLISHER_ID[i], host) && strstr(PUBLISHER_ID[i], s_port)) {
                pub_id = i;
                break;
            }
    }

    if (i == MAX_PUB) {
        /* Did not find anything, add a new entry */
        for (i=0; i<MAX_PUB; i++)
            if (PUBLISHER_ID[i] == NULL) {
                int len = strlen(host) + strlen(s_port) + 2;
                PUBLISHER_ID[i] = (char *) malloc (len * sizeof(char));
                if (PUBLISHER_ID[i] == NULL)
                    return AMS_ERROR_INSUFFICIENT_MEMORY;

                strcpy(PUBLISHER_ID[i], host);
                strcat(PUBLISHER_ID[i], "|");
                strcat(PUBLISHER_ID[i], s_port);
                pub_id = i;
                break;
            }
    }

    if (i == MAX_PUB)
        return AMS_ERROR_PUB_TBL_FULL; /* Maximum number of the publishers */

    return AMS_ERROR_NONE;
}


/*@
        AMSP_Get_publisher_id - Find the hostname and port number in the Communicators
        list and sets the current publisher id to the the index that was found.

        Input parameters:
+			host - Publisher's host name
+			port - Publisher's port number

@*/

int AMSP_Get_publisher_id(char *host, int port)
{
    int i, j, done = 0;
    char s_port[20];

    /* Convert port number to string */
    sprintf(s_port, "%d", port);

    for (i=0; i<MAX_PUB && !done; i++) {
        for (j=0; j<MAX_COMM && !done; j++) {
            if (COMM_LIST[i][j])
                if (strstr(COMM_LIST[i][j], host) && strstr(COMM_LIST[i][j], s_port)) {
                    pub_id = i;
                    done = 1;
                }
        }
    }


    if (i == MAX_PUB) {
        /* Did not find a publisher id for this communicator */
        return AMS_ERROR_PUBID_NOT_FOUND;
    }

    return AMS_ERROR_NONE;
}


#ifdef AMS_SERVER
/*@
        AMSP_Construct_Port_list - Find out communicator ports from other hosts when we publish
                              a new communicator

        Input parameters:
+			ams - ALICE Communicator being constructed
+			host_list - list of hosts to contact for ports

@*/

int AMSP_Construct_Port_list_AMS_Comm(AMS_Comm ams, char **host_list,
                                 int *port_list, char *name)
{
    int i, err, err1, cur_port, nbtry, ports_known = 0, t_flag;
    COMM_STRUCT *com_str;

    /* First, fill in our own port */
    err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    AMSP_LockMutexRead(&(com_str->pub_arg->pub_mutex));
    /* (wait for another thread to fill it in): */
    while(com_str->pub_arg->care_about_port != 2) {
        AMSP_CondWaitMutexRead(&(com_str->pub_arg->cond_port), &(com_str->pub_arg->pub_mutex), 0, &t_flag);
    }
    AMSP_RelMutexRead(&(com_str->pub_arg->pub_mutex));

    /* (now copy it into the host node list): */
    err = AMSP_Set_Comm_local_port(ams);
    CheckErr(err);
    
    if(host_list != NULL) {
        nbtry = 100;
        while(ports_known == 0 && nbtry) {
            ports_known = 1;
            for(i = 0; host_list[i]; ++i) {
                err = AMSP_Get_Comm_port(ams, i, &cur_port);
                CheckErr(err);
                if(cur_port <= 0) {
                    int sfd, answer;

                    if (AMSP_mem_net_open(host_list[i], AMS_SERVICE,
                                        port_list[i], &sfd, ams_msg) == AMS_ERROR_CONNECT_REFUSED ) {
                        ports_known = 0;
                        nbtry--;
                        AMS_Print("Still trying to contact publisher %s\n",
                               host_list[i]);
                    } else {
                        err = AMSP_send_rq(OP_COMM_PORT, sfd, name);
						CheckErr(err);

                        err1 = AMSP_fsm_loop(OP_COMM_PORT, sfd, &err, &answer);
                        AMSP_mem_net_close(sfd);
                        CheckErr(err1);
						CheckErr(err);

                        if(answer <= 0)
                            ports_known = 0;
                        else {
                            err = AMSP_Set_Comm_port(ams, i, answer);
                            CheckErr(err);
                        }
                    }
                }
            }
        }
    }
    else {
        /* Wait for the local port to become active */
        cur_port = -1;
        while(cur_port <= 0) {
            err = AMSP_Get_Comm_local_port(ams, &cur_port);
            CheckErr(err);
        }
    }

    err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    /* We're alive for client requests */
    AMSP_LockMutexWrite(&(com_str->pub_arg->pub_mutex));
    com_str->pub_arg->alive = 1;
    AMSP_RelMutexWrite(&(com_str->pub_arg->pub_mutex));

    return AMS_ERROR_NONE;              
}

int AMSP_Am_I_First_Host_AMS_Comm(AMS_Comm ams, int *answer)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    return AMSP_Am_I_First_Host_Comm_Struct(com_str, answer);
}

int AMSP_Who_Is_Next_Host_AMS_Comm(AMS_Comm ams, HOST_NODE **answer)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    return AMSP_Who_Is_Next_Host_Comm_Struct(com_str, answer);
}

int AMSP_Get_Nb_Proc_AMS_Comm(AMS_Comm ams, int *nb_proc)
{
    COMM_STRUCT *com_str;
    int err = AMSP_Get_Comm_Tbl(ams, &com_str);
    CheckErr(err);

    err = Validate(AMSP_Valid_Comm_Struct, com_str);
    CheckErr(err);

    return AMSP_Get_Nb_Proc_Comm_Struct(com_str, nb_proc);
}
#endif

#ifdef _DEBUG

/*@
        AMSP_Valid_Host_Node - Checks the validity of a Node

        Input Parameters:
+			host_node - Node to be validated

@*/
int AMSP_Valid_Host_Node(HOST_NODE *host_node)
{

    if (host_node == NULL)
        return AMS_ERROR_BAD_ARGUMENT;

    if (host_node->port != -1 && host_node->port < 0)
        return AMS_ERROR_BAD_ARGUMENT;

    return AMS_ERROR_NONE;
}

/*@
        AMSP_Valid_Comm_Struct - Checks the validity of a communication structure

        Input Parameters:
+			com_str - Communication structure to be validated

@*/
int AMSP_Valid_Comm_Struct(COMM_STRUCT *com_str)
{

    int err;
    if (com_str == NULL)
        return AMS_ERROR_BAD_ARGUMENT;

    if (com_str->nb_proc < 0)
        return AMS_ERROR_BAD_ARGUMENT;

    if (com_str->nb_mem < 0)
        return AMS_ERROR_BAD_ARGUMENT;

    err = Validate(AMSP_Valid_Host_Node, com_str->TailHostNode);
    CheckErr(err);

    err = Validate(AMSP_Valid_Host_Node, com_str->host_node);
    CheckErr(err);

    err = Validate(AMSP_Valid_Mem_Node, com_str->TailMemNode);
    CheckErr(err);

    err = Validate(AMSP_Valid_Mem_Node, com_str->mem_node);
    CheckErr(err);

    return AMS_ERROR_NONE;
}

#endif

