/*
 *  network handling for TCP/IP connection.
 */

#include        <signal.h>
#include        <stdlib.h>

#ifndef _WIN_ALICE
#include    <unistd.h>
#else
#include        <io.h>
#endif

#include    <stdio.h>
#include    <stdlib.h>
#include        <errno.h>

#include "ams.h"
#include "netlib.h"
#include "defs.h"
#ifdef AMS_SERVER

#ifdef _WIN_ALICE

#define bzero(s, n) memset((void *)s, 0, n)
#define bcopy(s1, s2, n)  memcpy(s2, s1, n)

#endif

#endif

#include        "network.h"
#include        "netdefs.h"

extern int      errno;

#ifndef AMS_CLIENT
#ifndef AMS_SERVER
either AMS_CLIENT or AMS_SERVER must be defined
#endif
#endif

char    openhost[MAXHOSTNAMELEN] = { 0 };       /* remember host's name */

extern int                      traceflag;      /* ALICE trace variable */


/*
 * Close the network connection.  Used by client and server.
 */

int AMSP_mem_net_close(int sfd)
{

#ifdef _WIN_ALICE
    closesocket(sfd);
#else
    close(sfd);
#endif

    return 0;
}

/*
 * Send a record to the other end.  Used by client and server.
 * With a stream socket we have to preface each record with its length,
 * We encode the length as a 2-byte integer in network byte order.
 */

int AMSP_mem_net_send(buff, len, sfd)
     char       *buff;
     int        len, sfd;
{
  
    int rc, err;
    int templen, tempver;
	char mesg[255];

    templen = htonl(len);
    rc = AMSP_writen(sfd, (char *) &templen, sizeof(int));
	if (rc == 0)
		return AMS_ERROR_CONNECT_TERMINATED;

    if (rc != sizeof(int))
        AMSP_err_quit("writen error of length prefix %s", AMSP_host_err_str(mesg, &err));

    tempver = htonl(AMS_MEMORY_NET_VERSION);
    rc = AMSP_writen(sfd, (char *) &tempver, sizeof(int));
	if (rc == 0)
		return AMS_ERROR_CONNECT_TERMINATED;

    if (rc != sizeof(int))
        AMSP_err_quit("writen error of version prefix %s", AMSP_host_err_str(mesg, &err));

    rc = AMSP_writen(sfd, buff, len);
	if (rc == 0)
		return AMS_ERROR_CONNECT_TERMINATED;

    if (rc != len)
        AMSP_err_quit("writen error %s", AMSP_host_err_str(mesg, &err));

    return 0;
}

/*
 * Receive a record from the other end.  Used by client and server.
 */

int AMSP_mem_net_recv(buff, nbytes, sfd)
     char **buff;
     int  *nbytes, sfd;
{
    int err, templen;
	char mesg[255];

#if !defined(_WIN_ALICE) && defined(AMS_SERVER)
    extern void AMSP_Pipe(int);
    signal(SIGPIPE, AMSP_Pipe);
#endif

again1:
    if ( (*nbytes = AMSP_readn(sfd, (char *) &templen, sizeof(int))) < 0) {
        if (errno == EINTR) {
            errno = 0;          /* assume SIGCLD */
            goto again1;
        }
        AMSP_err_sys("AMSP_mem_net_recv: readn failed");
    }
	if (*nbytes == 0)
		return AMS_ERROR_CONNECT_TERMINATED;

    if (*nbytes != sizeof(int))
        AMSP_err_quit("AMSP_mem_net_recv: Expecting 4 bytes op code, got %d", *nbytes);

    templen = ntohl(templen) + INT_SIZE;   /* #bytes that follow; includes version prefix */
    if (templen < 0 || templen == 0)
        AMSP_err_quit("AMSP_mem_net_recv: Invalid message length %d", templen);

    *buff = (char *) malloc(sizeof(char) * templen);
    if (*buff == NULL)
        AMSP_err_quit("AMSP_mem_net_recv: Can't allocate memory for incoming message");

again2:
    if ( (*nbytes = AMSP_readn(sfd, *buff, templen)) < 0) {
        if (errno == EINTR) {
            errno = 0;          /* assume SIGCLD */
            goto again2;
        }
        AMSP_err_quit("AMSP_readn error %s", AMSP_host_err_str(mesg, &err));
    }
	if (*nbytes == 0)
		return AMS_ERROR_CONNECT_TERMINATED;

    if (*nbytes != templen)
        AMSP_err_quit("error in readn. Expecting %d, received %d ", nbytes, templen);

    return AMS_ERROR_NONE;
}

#ifdef  AMS_SERVER

struct sockaddr_in              tcp_cli_addr;   /* set by accept() */

/*
 * Initialize the network connection for the server,
 * if the port number is -1, get port# from the system
 */

int AMSP_mem_net_init(service, care_about_port, port, sfd, errmsg)
     char       *service, *errmsg;      /* the name of the service we provide */
     int        care_about_port, *port, *sfd;

{
    struct servent      *sp;
    struct sockaddr_in  tcp_srv_addr;
    int err;
#ifndef _WIN_ALICE
int optval;
#endif

#if defined(RS6000)
    size_t sinlen = sizeof(struct sockaddr_in);
#else
    int sinlen = sizeof(struct sockaddr_in);
#endif

    /*
     * We weren't started by a master daemon.
     * We have to create a socket ourselves and bind our well-known
     * address to it.
     */

    bzero((char *) &tcp_srv_addr, sizeof(tcp_srv_addr));
    tcp_srv_addr.sin_family      = AF_INET;
    tcp_srv_addr.sin_addr.s_addr = htonl(INADDR_ANY);

    if (service != NULL) {
        if ( (sp = getservbyname(service, "tcp")) == NULL)
            AMSP_err_quit("AMSP_mem_net_init: unknown service: %s/tcp", service);

        if (*port > 0)
            tcp_srv_addr.sin_port = htons((short)(*port));
        /* caller's value */
        else
            tcp_srv_addr.sin_port = sp->s_port;
        /* service's value */
    } else {
        if (*port == -1)
            *port = 0;

        if (*port < 0) {
            AMSP_err_ret(errmsg, "AMSP_mem_net_init: must specify a valid port");
            return(-1);
        }
        tcp_srv_addr.sin_port = htons((short)(*port));
    }

    /*
     * Create the socket and Bind our local address so that any
     * client can send to us.
     */

    if ( (*sfd = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        AMSP_err_dump("AMSP_mem_net_init: can't create stream socket");

#ifndef _WIN_ALICE
    /* On NT the port will be "re-used" even though there is another */
    /* using the current port. */
    /* Set socket option to reuse an address after prompt restart */
    optval = 1; /* Turn on the option */
    err = setsockopt(*sfd, SOL_SOCKET, SO_REUSEADDR, (char *)&optval, sizeof(optval));
    if (err < 0)
        AMSP_err_sys(" SO_REUSEADDR in setsockopt call ");
#endif
    
    while ((err = bind(*sfd, (struct sockaddr *) &tcp_srv_addr,
                  sizeof(tcp_srv_addr))) < 0) {

        if(care_about_port) {
            perror(" Can't bind socket");
            AMSP_err_dump("AMSP_mem_net_init: can't bind local address");
        } else {
            /* Try another port */
            *port = 0;
            tcp_srv_addr.sin_port = htons((short)(*port));
        }
    }

    /* Get back the port number back*/
    if (*port == 0) {
        if ((err = getsockname( *sfd, (struct sockaddr *) &tcp_srv_addr,
                                &sinlen)) < 0)
            AMSP_err_dump("getsockname: can't get assigned port number");
        *port = ntohs(tcp_srv_addr.sin_port);
    }

    /*
     * And set the listen parameter, telling the system that we're
     * ready  to accept incoming connection requests.
     */

    err=listen(*sfd, 5);
    return err;
}

/*
 * The server waits for client request on a known port
 * inetd is a flag indicating whether we were started by the deamon
 * alive is a flag used to indicate whether the server is being shutdown
 * In case of shutdown we should not accept any new connection. Note that
 * the alive lag will work only in a multithreaded environment or shared
 * memory.
 */

int AMSP_mem_net_accept(int inetd, int sfd, void *arg)
{
    int newsockfd, err, i = 0, done = 0;
#if defined (RS6000)
    size_t clilen;
#else
    int clilen;
#endif

    struct sockaddr_in          tcp_cli_addr;

    extern int AMSP_ProcessRequest (void *, int );
    extern void AMSP_Term(int);

    signal(SIGTERM, AMSP_Term);

    while (!done) {
        clilen = sizeof(tcp_cli_addr);
        newsockfd = accept(sfd, (struct sockaddr *) &tcp_cli_addr, &clilen);
        done = 1;
        if (newsockfd < 0) {
            if (errno == EINTR) {
                errno = 0;
                done = 0;
            }
            else {
                AMSP_err_sys("accept failed. system error: ");
                AMSP_err_dump("accept error errno=%d newsockfd=%d",
                         errno, newsockfd);
            }
        }
    }

    /*
     * Create a Child Process or a Worker Thread to handle the new
     * request. We are passing a pointer to the new socket descriptor
     * in case it is a Worker Thread.
     */

    err = AMSP_ProcessRequest(arg, newsockfd);

    if (err == 1)
        /* Server being shutdown */
        return 0;
    else
        /* The main thread/process returns to listen for a new request */
        return 1;

}

/*
 * Signal Handlers
 */

void AMSP_Term(int sfd)
{
    AMSP_mem_net_close(sfd);
    exit(0);
}

#endif
/*
 * Open the network connection. Used by both ClIENT and AMS_SERVER
 */

int AMSP_mem_net_open(char *host, char *service, int port, int *sfd, char *errmsg)
{
    struct sockaddr_in  tcp_srv_addr;
    struct servent tcp_srv_info;
	int err;

    *sfd = AMSP_tcp_open(host, service, port, &tcp_srv_addr, &tcp_srv_info, &err, errmsg);
	if (*sfd < 0) {

#ifndef _WIN_ALICE
		if (err == ECONNREFUSED || err == ECONNRESET)
			return AMS_ERROR_CONNECT_REFUSED;
#else
		if (err == WSAECONNREFUSED || err == WSAECONNRESET)
			return AMS_ERROR_CONNECT_REFUSED;
#endif
		AMSP_err_quit(" Fatal error %d, in connecting to the other side \n", err);
	}

	return AMS_ERROR_NONE;
}

#ifndef _WIN_ALICE
/* Don't know how to take care of broken pipes in Windows */
void AMSP_Pipe(int sig)
{
 
#ifdef _MT
    signal(SIGPIPE, SIG_IGN);
#else
    signal(SIGPIPE, SIG_DFL);
#endif
             
}
#endif

